from pathlib import Path
import urllib.request
import subprocess, json, sys


def upload_package(package_name:str, current_version:str) -> None:
    # TODO: look into `class UploadCommand(setuptools.Command)` from <https://github.com/kennethreitz/setup.py/blob/master/setup.py#L49>
    package_name = package_name.lower()

    # Make sure there's no unstaged changes
    git_workdir_returncode = subprocess.run('git diff-files --quiet'.split()).returncode
    assert git_workdir_returncode in [0,1]
    if git_workdir_returncode == 1:
        print('=> git workdir has changes')
        print('=> please either revert or stage them')
        sys.exit(1)

    # If the local version is the same as the PyPI version, increment it.
    pypi_url = f'https://pypi.python.org/pypi/{package_name}/json'
    latest_version = json.loads(urllib.request.urlopen(pypi_url).read())['info']['version']
    # Note: it takes pypi a minute to update the API, so this can be wrong.
    if latest_version == current_version:
        new_version = next_version(current_version)
        print(f'=> autoincrementing version {current_version} -> {new_version}')
        Path(f'{package_name}/version.py').write_text(f"version = '{new_version}'\n")
        current_version = new_version
        subprocess.run(['git','stage',f'{package_name}/version.py'], check=True)

    # Commit any staged changes
    git_index_returncode = subprocess.run('git diff-index --quiet --cached HEAD'.split()).returncode
    assert git_index_returncode in [0,1]
    if git_index_returncode == 1:
        print('=> git index has changes; committing them')
        subprocess.run(['git','commit','-m',current_version], check=True)

    # Clean and repopulate ./dist/*
    if Path('dist').exists() and list(Path('dist').iterdir()):
        # Double check that we are where we think we are
        setuppy = Path('dist').absolute().parent / 'setup.py'
        assert setuppy.is_file() and package_name in setuppy.read_text().lower()
        for child in Path('dist').absolute().iterdir():
            assert child.name.lower().startswith(f'{package_name}-'), child
            print('=> unlinking', child)
            child.unlink()
    subprocess.run('python3 setup.py sdist bdist_wheel'.split(), check=True)

    # Upload
    if not Path('~/.pypirc').expanduser().exists():
        print('=> warning: you need a ~/.pypirc')
    try: subprocess.run(['twine','--version'], check=True)
    except Exception: print('Run `pip3 install twine` and try again'); sys.exit(1)
    subprocess.run('twine upload dist/*'.split(), check=True)

    if git_index_returncode == 1:
        print('=> Now do `git push`.')

def next_version(version:str) -> str:
    new_version_parts = version.split('.')
    new_version_parts[2] = str(1+int(new_version_parts[2]))
    return '.'.join(new_version_parts)
assert next_version('1.1.5') == '1.1.6'



import time, random, sys, subprocess
from pathlib import Path
from functools import wraps

def run(filepath:str = '', make_cache:bool = True, run_rarely:bool = False) -> None:
    if run_rarely:
        seconds_since_last_change = time.time() - Path(filepath).stat().st_mtime
        if seconds_since_last_change > 30 and random.random() > 0.01:
            return  # Don't run
    run_flake8(filepath)
    run_mypy(filepath, make_cache=make_cache)

def run_flake8(filepath:str = '') -> None:
    p = subprocess.run(['flake8', '--show-source', '--ignore=E501,E302,E251,E701,E226,E305,E225,E261,E231,E301,E306,E402,E704,E265,E201,E202,E303,E124,E241,E127,E266,E221,E126,E129,F811,E222,E401,E702,E203,E116,E228,W504,W293,B007,W391,F401,W292,E227,E128', filepath])
    if p.returncode != 0: sys.exit(1)

def run_mypy(filepath:str = '', make_cache:bool = True) -> None:
    cmd = ['mypy', '--pretty', '--ignore-missing-imports']
    if not make_cache: cmd.append('--cache-dir=/dev/null')
    if filepath: cmd.append(filepath)
    p = subprocess.run(cmd)
    if p.returncode != 0: sys.exit(1)


def get_size(obj, seen:set = None) -> int:
    """Recursively calculates bytes of RAM taken by object"""
    # From https://code.activestate.com/recipes/577504/ and https://github.com/bosswissam/pysize/blob/master/pysize.py
    if seen is None: seen = set()
    size = sys.getsizeof(obj)
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    seen.add(obj_id)  # Mark as seen *before* recursing to handle self-referential objects

    if isinstance(obj, dict):
        size += sum(get_size(v, seen) for v in obj.values())
        size += sum(get_size(k, seen) for k in obj.keys())
    elif hasattr(obj, '__dict__'):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum(get_size(i, seen) for i in obj)

    if hasattr(obj, '__slots__'):  # obj can have both __slots__ and __dict__
        size += sum(get_size(getattr(obj, s), seen) for s in obj.__slots__ if hasattr(obj, s))

    return size




import re, json, urllib.request

def check_pkg(pkg, opt, version, line=None):
    '''
    pkg is like "requests"
    opt is like "[security]" or ""
    version is like ">=4.0"
    line is for debugging
    '''
    if opt is None: opt=''
    if version is None: version=''
    try:
        j = json.loads(urllib.request.urlopen('https://pypi.org/pypi/{}/json'.format(pkg)).read())
        latest = j['info']['version']
        v = version.lstrip('~=>')
        update_str = '' if latest == v or latest==v+'.0' or latest.startswith(v) else '>>'
        print('{:<2} {:20} {:10} {:10}'.format(update_str, pkg+opt, version, latest))
    except Exception:
        raise Exception([pkg, opt, version, line])

def check_file(filepath):
    with open(filepath) as f:
        for line in f:
            m = re.match(r'''^\s*'?([-a-zA-Z]+)(\[[a-zA-Z]+\])?([~<>=]{2}[0-9a-zA-Z\.]+)?'?,?\s*$''', line)
            if m:
                pkg, opt, version = m.group(1), m.group(2), m.group(3)
                check_pkg(pkg, opt, version, line)



from typing import Any, cast, Callable, TypeVar, Iterator
import shelve, json, os.path


def _jsonify(obj: Any) -> str:
    return json.dumps(obj, default=_jsonify_default)
def _jsonify_default(obj: Any) -> Any:
    try:
        import pandas as pd
    except ImportError:
        pass
    else:
        if isinstance(obj, pd.DataFrame):
            return {'type=pandas.DataFrame': obj.to_json()}
    raise TypeError('Object of type {obj.__class__.__name__} is not JSON serializable!'.format(**locals()))


F = TypeVar('F', bound=Callable[..., Any])

def shelve_cache(func:F) -> F:
    filename = '.cache-{func.__name__}.shelve'.format(**locals())
    def wrapper(*args, **kwargs):
        key = _jsonify([args, kwargs])
        # with shelve.open(filename) as shelf:
        #     if key not in shelf: shelf[key] = func(*args, **kwargs)
        #     return shelf[key]
        with shelve.open(filename) as shelf:
            if key in shelf: return shelf[key]
        # release the file to allow other processes to use it
        # this function is meant for slow functions, so this extra closing and opening isn't significant
        ret = func(*args, **kwargs)
        with shelve.open(filename) as shelf: shelf[key] = ret
        return ret
    return cast(F, wrapper)

def cache_by_id(func:F) -> F:
    cache = {}
    def wrapper(*args, **kwargs):
        key = _jsonify([[id(arg) for arg in args], {k:id(v) for k,v in kwargs}])
        if key not in cache: cache[key] = func(*args, **kwargs)
        return cache[key]
    return cast(F, wrapper)


T = TypeVar('T')
F2 = TypeVar('F2', bound=Callable[[], Iterator[T]])

def cached_generator(record_maker:Callable[[Any],T]=lambda x:x) -> Callable[[F2],F2]:
    '''
    Caches a generator into jsonlines format.
    Deserialization is done by `record_maker(json.loads(line))`.
    Doesn't track args, so don't use any.
    '''
    def decorator(func:F2) -> F2:
        cache_fpath = os.path.join('.cache-{func.__name__}.jsonlines'.format(**locals()))
        def wrapper(*args, **kwargs):
            if args or kwargs: raise Exception('cached_generator() cannot handle args or kwargs')
            if not os.path.exists(cache_fpath):
                with open(cache_fpath+'.tmp', 'wt') as f:
                    for r in func():
                        f.write(json.dumps(r) + '\n')
                        yield r
                os.rename(cache_fpath+'.tmp', cache_fpath)
            else:
                with open(cache_fpath, 'rt') as f:
                    for line in f:
                        yield record_maker(json.loads(line))
        return cast(F2, wrapper)
    return decorator

def cached_generator_json(func:F2) -> F2:
    '''
    Caches a generator into jsonlines format.
    Deserialization is done by `json.loads(line)`.
    Doesn't track args, so don't use any.
    '''
    cache_fpath = os.path.join('.cache-{func.__name__}.jsonlines'.format(**locals()))
    def wrapper(*args, **kwargs):
        if args or kwargs: raise Exception('cached_generator() cannot handle args or kwargs')
        if not os.path.exists(cache_fpath):
            with open(cache_fpath+'.tmp', 'wt') as f:
                for r in func():
                    f.write(json.dumps(r) + '\n')
                    yield r
            os.rename(cache_fpath+'.tmp', cache_fpath)
        else:
            with open(cache_fpath, 'rt') as f:
                for line in f:
                    yield json.loads(line)
    return cast(F2, wrapper)



import math

def round_sig(x, digits):
    if x == 0:
        return 0
    elif abs(x) == math.inf or math.isnan(x):
        raise ValueError("Cannot round infinity or NaN")
    else:
        log = math.log10(abs(x))
        digits_above_zero = int(math.floor(log))
        return round(x, digits - 1 - digits_above_zero)
assert round_sig(0.00123, 2) == 0.0012
assert round_sig(0.00100, 2) == 0.001
assert round_sig(1.59e-10, 2) == 1.6e-10




import sys, html, urllib, re, os


def status_code_server(environ, start_response):
    # TODO: query all status codes using [urllib.request.urlopen, requests.get, requests.get.raise_for_status] to compare error-handling
    #  - results should be: (urlopen / requests.get*)
    #    - can't connect => URLError / requests.exceptions.ConnectionError
    #    - 000-100 => BadStatusLine / requests.exceptions.ConnectionError
    #    - 101-199 => HTTPError / ok
    #    - 200-299 => ok
    #    - 301-303,307 with valid "Location:" header => respond for new location
    #    - 300-399 otherwise => HTTPError / ok
    #    - 400-599 => HTTPError / ok (but raise_for_status raises requests.exceptions.HTTPError)
    #    - 600-999 => HTTPError / ok
    headers = [('Content-type', 'text/plain')]
    path = environ.get('PATH_INFO','')

    m = re.match(r'^/([0-9]{3})/([0-9]{3})$', path)
    if m:
        status = m.group(1) + ' WAT'
        headers.append(('Location', '/{}'.format(m.group(2))))
        ret = 'following xxx -> xxx path for {}\n'.format(path).encode('utf8')
        headers.append(('Content-Length', str(len(ret))))
        start_response(status, headers)
        return [ret]

    m = re.match(r'^/([0-9]{3})$', path)
    if m:
        status = m.group(1) + ' WAT'
        ret = 'following xxx path for {}\n'.format(path).encode('utf8')
        headers.append(('Content-Length', str(len(ret))))
        start_response(status, headers)
        return [ret]

    raise Exception('bad url: {path}'.format(**locals()))


def make_redirect_server(target_base_url):
    assert re.match(r'https?://(?:[-a-z0-9]+\.)+[-a-z0-9]+(?:/.*)?', target_base_url)
    def redirect_server(environ, start_response):
        headers = [('Content-type', 'text/plain')]
        path = environ.get('PATH_INFO','')
        if environ.get('QUERY_STRING'):
            path += '?' + environ.get('QUERY_STRING')
        if re.match(r'^[-/a-zA-Z0-9%\.+~_=:\?&]*$', path):
            status = '302 Found'
            headers.append(('Location', '{}{}'.format(target_base_url, path)))
            ret = 'redirecting to {}{}\n'.format(target_base_url, path)
        else:
            status = '404 Not Found'
            ret = 'URL not permitted\n'
        start_response(status, headers); return [ret.encode('utf8')]
    return redirect_server

def directory_server(environ, start_response):
    # TODO: just implement a simple version by hand
    def normalize_path(path):
        # This code is taken from http.server.SimpleHTTPRequestHandler.translate_path()
        import urllib.parse, posixpath
        path = path.split('?', 1)[0]
        path = path.split('#', 1)[0]
        path = urllib.parse.unquote(path)
        path = posixpath.normpath(path)
        words = [word for word in path.split('/') if word]
        path = os.getcwd()
        for word in words:
            if os.path.dirname(word) or word in (os.curdir, os.pardir):
                # ignore components like `.`, `..`, (and also somehow directories?)
                continue
            path = os.path.join(path, word)
        return path
    path = normalize_path(environ.get('PATH_INFO', ''))



def magic_directory_server(environ, start_response):
    # This is like `python3 -m http.server` (http.server.SimpleHTTPRequestHandler).
    # `SimpleHTTPRequestHandler` chooses mimetype using file extension (mimetypes.types_map)
    # This server instead uses mime-guessing
    # I have a solid implementation in `MimeSniffingHTTPRequestHandler` but I'm not sure how to convert it to WSGI/gunicorn
    #    It extends `http.server.SimpleHTTPRequestHandler`
    #       which extends `http.server.BaseHTTPRequestHandler`
    #           which extends `socketserver.StreamRequestHandler`.
    #    It's run using `http.server.test(HandlerClass=MimeSniffingHTTPRequestHandler, ServerClass=http.server.HTTPServer)`
    #       which uses `http.server.HTTPServer(('localhost', 8001), MimeSniffingHTTPRequestHandler).serve_forever()`
    #          which extends `socketserver.TCPServer`
    def normalize_path(path):
        # This code is taken from http.server.SimpleHTTPRequestHandler.translate_path()
        import urllib.parse, posixpath
        path = path.split('?', 1)[0]
        path = path.split('#', 1)[0]
        has_trailing_slash = path.rstrip().endswith('/')
        path = urllib.parse.unquote(path)
        path = posixpath.normpath(path)
        words = [word for word in path.split('/') if word]
        path = os.getcwd()
        for word in words:
            if os.path.dirname(word) or word in (os.curdir, os.pardir):
                # ignore components like `.`, `..`, (and also somehow directories?)
                continue
            path = os.path.join(path, word)
        if has_trailing_slash:
            path += '/'
        return path
    def generate_directory_listing(path):
        enc = sys.getfilesystemencoding()
        try:
            lst = os.listdir(path)
        except OSError:
            return 'Failed to list directory'.encode(enc), enc
        lst.sort(key=lambda a: a.lower())
        r = []
        displaypath = html.escape(urllib.parse.unquote(path), quote=False)
        title = 'Directory listing for %s' % displaypath
        r.append('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" '
                 '"http://www.w3.org/TR/html4/strict.dtd">')
        r.append('<html>\n<head>')
        r.append('<meta http-equiv="Content-Type" '
                 'content="text/html; charset=%s">' % enc)
        r.append('<title>%s</title>\n</head>' % title)
        r.append('<body>\n<h1>%s</h1>' % title)
        r.append('<hr>\n<ul>')
        for name in lst:
            displayname = linkname = name
            fullname = os.path.join(path, name)
            if os.path.isdir(fullname):
                displayname = linkname = name + '/'
            if os.path.islink(fullname):
                displayname = name + '@'
            r.append('<li><a href="%s">%s</a></li>'
                     % (urllib.parse.quote(linkname),
                        html.escape(displayname, quote=False)))
        r.append('</ul>\n<hr>\n</body>\n</html>\n')
        return '\n'.join(r).encode(enc), enc
    def guess_content_type(fpath):
        # TODO: use `http.server.SimpleHTTPRequestHandler.guess_type` method
        import magic
        return magic.from_file(fpath, mime=True)
    path = normalize_path(environ.get('PATH_INFO', ''))
    if os.path.isdir(path):
        parts = urllib.parse.urlsplit(environ.get('PATH_INFO',''))
        if not parts.path.endswith('/'):
            # redirect browser like apache2 does
            new_parts = (parts[0], parts[1], parts[2] + '/', parts[3], parts[4])
            new_url = urllib.parse.urlunstplit(new_parts)
            start_response('301 MOVEDPERMANENTLY', [('Location', new_url)])
            return b''
        else:
            data, enc = generate_directory_listing(path)
            start_response('200 OK', [('Content-type', 'text/html; charset=%s' % enc)
                                      ('Content-Length', str(len(data)))])
            return data
    ctype = guess_content_type(path)
    try:
        f = open(path, 'rb')
    except OSError:
        start_response('404 NOTFOUND', [('Connection', 'close')])
    # Note: work-in-progress

def serve(app, port=5000):
    from .http_utils import run_gunicorn
    try:
        run_gunicorn(app, port=port, use_reloader=False)
    except KeyboardInterrupt:
        pass




def use_ipdb():
    # TODO: make `import kpa.terminal_utils.use_ipdb` set `sys.excepthook` without `use_ipdb()`?
    #       - then would all of kpa.terminal_utils.* need to be modules, rather than functions?
    # TODO: figure out how to use `python3 -m ipdb run.py` or `ipdb3 run.py`. Are they broken on py3?
    import sys
    def excepthook(*args, **kwargs):
        from IPython.core import ultratb
        sys.excepthook = ultratb.FormattedTB(mode='Verbose', color_scheme='Linux', call_pdb=1)
        return sys.excepthook(*args, **kwargs)
    sys.excepthook = excepthook


def ignore_sigpipe():
    import signal
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)


def termcolor(text:str, fg:int = None, bg:int = None, under:bool = False) -> str:
    # TODO: look at <https://github.com/kennethreitz/crayons/blob/master/crayons.py>
    from subprocess import check_output
    def get_code(args):
        if not hasattr(termcolor,'_cache'): termcolor._cache = {}
        if termcolor._cache.get(args, None) is None: termcolor._cache[args] = check_output('tput {args}'.format(args=args).split()).decode()
        return termcolor._cache[args]
    return ((get_code('setaf {fg}'.format(fg=fg)) if fg else '') +
            (get_code('setab {bg}'.format(bg=bg)) if bg else '') +
            (get_code('smul') if under else '') +
            text + get_code('sgr0'))
termcolor.BG_RED = 1
termcolor.BG_GREEN = 2
termcolor.BG_YELLOW = 3
termcolor.BG_BLUE = 4
termcolor.BG_PINK = 5
termcolor.BG_GRAY = 6
termcolor.BG_WHITE = 14

class TerminalLineWrapper:
    # TODO: make these classmethods
    '.write_text(text) wraps long lines using an eight-space indent and returns the result'
    def __init__(self):
        import shutil
        self.cols = shutil.get_terminal_size().columns
    def write_text(self, text):
        for line in text.split('\n'):
            self._write_line(line)
    def _write_line(self, text):
        ret = ''
        x = self._how_many_chars_fit_in_width(text, self.cols)
        ret += (text[:x])
        while x < len(text):
            text = text[x:]
            x = self._how_many_chars_fit_in_width(text, self.cols - 8)
            ret += '\n' + ' '*8 + text[:x]
        return ret
    def _how_many_chars_fit_in_width(self, s, width):
        for i, c in enumerate(s):
            width -= self._width_of_char(c)
            if width < 0: return i
        return len(s)
    def _width_of_char(self, c):
        import unicodedata
        ## A: ambiguous (1 in iterm)
        ## F: full width (2)
        ## H: halfwidth (1)
        ## N : not asian (1)
        ## Na: narrow (1)
        ## W: wide (2)
        u_eaw = unicodedata.east_asian_width(c)
        if u_eaw in ('H','N','Na','A'): return 1
        elif u_eaw in ('F','W'): return 2
        else: raise Exception(ord(c))



#!/usr/bin/env python3
# to install locally: `pip install -e .`
# to install latest from pypi: `pip3 install --upgrade --upgrade-strategy eager --no-cache-dir kpa`
# to publish: `./setup.py publish`
# to update deps: `kpa pip-find-updates`

from setuptools import setup
import importlib
import sys


def load_module_by_path(module_name, filepath):
    module = importlib.util.module_from_spec(importlib.util.spec_from_file_location(module_name, filepath))
    module.__spec__.loader.exec_module(module)
    return module
version = load_module_by_path('kpa.version', 'kpa/version.py').version


if sys.argv[-1] in ['publish', 'pub']:
    pypi_utils = load_module_by_path('kpa.pypi_utils', 'kpa/pypi_utils.py')
    pypi_utils.upload_package(package_name='Kpa', current_version=version)
    sys.exit(0)


setup(
    name='Kpa',
    version=version,
    description="<forthcoming>",
    author="Peter VandeHaar",
    author_email="pjvandehaar@gmail.com",
    url="https://github.com/pjvandehaar/kpa",
    classifiers=[
        'Programming Language :: Python :: 3 :: Only',
        'Operating System :: Unix',
    ],

    package_data={'kpa': ['py.typed']},  # tells mypy this has types
    packages=['kpa'],
    entry_points={'console_scripts': [
        'kpa=kpa.command_line:main',
    ]},
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.4",
    setup_requires=[
        'pytest-runner~=4.2',
    ],
    install_requires=[
        'boltons~=20.2',
    ],
    tests_require=[
        'pytest~=4.0',
    ],
)



import shutil
import bs4

def _not_empty(iterable):
    try:
        next(iterable)
        return True
    except StopIteration:
        return False

def limit_width(text, maxwidth):
    if maxwidth is None: maxwidth = shutil.get_terminal_size().columns
    if maxwidth == 0: maxwidth = 1_000_000_000
    return '\n'.join(line[:maxwidth] for line in text.split('\n'))


def repr_node(node):
    if isinstance(node, str): return repr(node)
    ret = '<{node.name}'.format(node=node)
    if 'id' in node.attrs: ret += ' id="{}"'.format(node.attrs["id"])
    if 'class' in node.attrs: ret += ' class="{}"'.format(" ".join(node.attrs["class"]))
    for k,v in node.attrs.items():
        if k not in ('id','class'): ret += ' {k}="{v}"'.format(**locals())
    return ret + '>'


def render_tree(node, depth=1):
    ret = repr_node(node)
    if isinstance(node, bs4.element.Tag) and _not_empty(node.children):
        if depth >= 1:
            for child in node.children:
                for line in render_tree(child, depth=depth-1).split('\n'): ret += '\n   ' + line
        else:
            num_descendants = sum(1 for _ in node.descendants)
            ret += '\n   ... {num_descendants}'.format(**locals())
    return ret
def print_tree(node, depth=1, maxwidth=None):
    print(limit_width(render_tree(node, depth), maxwidth))

def render_ancestors(node):
    if node.parent is None: return ''
    return render_ancestors(node.parent) + '\n' + repr_node(node)
def get_leaf_nodes_containing_text(soup, text):
    for node in soup.descendants:
        if isinstance(node, str) and not isinstance(node, bs4.element.Doctype) and text in node:
            yield node
def find_paths_to_text(soup, text, maxwidth=None):
    for node in soup.descendants:
        if isinstance(node, str) and not isinstance(node, bs4.element.Doctype) and text in node:
            print(limit_width(render_ancestors(node), maxwidth))
            print()

if __name__ == '__main__':
    from bs4 import BeautifulSoup
    from kpa.http_utils import cached_get
    soup = BeautifulSoup(cached_get('https://google.com').text, 'lxml')
    print_tree(soup.select_one('html'), depth=3)



import functools
from typing import List, Callable, Any, Iterable, TypeVar, Tuple, Dict, cast, Iterator

ReturnType = TypeVar('ReturnType')
F = TypeVar('F', bound=Callable[...,ReturnType])
T = TypeVar('T')
T2 = TypeVar('T2')

def assign(func: Callable[[], T]) -> T: return func()
def assign_list(func:Callable[..., Iterator[T]]) -> List[T]: return list(func()) # TODO: prefer `@assign @list_from_iter`, which also allows cacheing
def assign_dict(func:Callable[..., Iterator[Tuple[T,T2]]]) -> Dict[T,T2]: return dict(func())

def only_once(func: F) -> F:
    already_ran = [False]
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not already_ran[0]:
            already_ran[0] = True
            return func(*args, **kwargs)
    return cast(F, wrapper)

# TODO: this isn't checking that their args/kwargs are the same.
def list_from_iter(func: Callable[..., Iterable[T]]) -> Callable[..., List[T]]:
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return list(func(*args, **kwargs))
    return cast(Callable[..., List[T]], wrapper)

def dict_from_iter(func: Callable[..., Iterable[Tuple[T, T2]]]) -> Callable[..., Dict[T, T2]]:
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return dict(func(*args, **kwargs))
    return cast(Callable[..., Dict[T,T2]], wrapper)

# Too complex to typecheck
def apply_to_result(*result_wrapper_funcs):
    def wrap(func):
        @functools.wraps(func)
        def f(*args, **kwargs) -> Any:
            result = func(*args, **kwargs)
            for result_wrapper_func in result_wrapper_funcs:
                result = result_wrapper_func(result)
            return result
        return f
    return wrap


def show(f:F) -> F:
    @functools.wraps(f)
    def wrapper(*args:object, **kwargs:object):
        ret: Any = f(*args, **kwargs)
        params: List[str] = []
        for arg in args: params.append(repr(arg))
        for k,v in kwargs.items(): params.append(f'{k}={repr(v)}')
        print(f'{f.__name__}({", ".join(params)}) = {ret}')
        return ret
    return cast(F, wrapper)
